#
# Ben Tribelhorn
#
# main.py
#

import RobotCanvas
import Tkinter
import time
import threading
import random
import Queue
import math
import sys
import socket

import roombaSim
import sensorSimulator
from mcl import MCL
import controller
from path import roombaHistory
from slam import SLAM, feature

NUM_PARTS = 0
NUM_PATHS = 1
MAP_NAME = 'mapMaze.txt'

HOST = '127.0.0.1'			# localhost
#HOST = '10.0.0.1'			# other laptop
PORT = 8000					# The port used by vision

BTPORT = '/dev/cu.RooTooth-COM0-1'

class GuiPart:
    def __init__(self, master, queue):
        self.queue = queue
        self.master = master
        #
        self.rf = RobotCanvas.RobotFrame(master)
        # right now, maps have to be added before robots
        self.mapname = MAP_NAME
        #self.rf.canv.loadMap(self.mapname)
        self.parts = []
        for i in range(NUM_PARTS):
            self.parts.append(RobotCanvas.MovableCircle(self.rf.canv, 0,0,0, 'black', 5.0))
        self.simrobot = RobotCanvas.MovableCircle( self.rf.canv, colorstr='red' , radius=16.5)
        self.hist = RobotCanvas.MovablePath(self.rf.canv, [], colorstr='blue')
        self.mostProbableSLAMPath = RobotCanvas.MovableSLAMPath(self.rf.canv, [], colorstr='orange')
        
    def shutdown(self):
        self.rf.quit()
        self.master.destroy()
        sys.exit(0)

    def processIncoming(self):
        """
        Handle all the messages currently in the queue (if any).
        """
        self.rf.canv.redrawRemote()
        """
        while self.queue.qsize():
            try:
                list = self.queue.get(0)
                #execute function with 0 or more parameters
                list[0](*list[1:])
            except Queue.Empty:
                pass
        """
        self.master.after(20, self.processIncoming) # 50 hz

class ThreadedClient:
    """
    Launch the main part of the GUI and the worker thread. periodicCall and
    endApplication could reside in the GUI part, but putting them here
    means that you have all the thread controls in a single place.
    """
    def __init__(self, master):
        """
        Start the GUI and the asynchronous threads. We are in the main
        (original) thread of the application, which will later be used by
        the GUI. We spawn a new thread for the worker.
        """
        self.master = master
        
        # Create the queue
        self.queue = Queue.Queue()
        # Set up the GUI part
        self.gui = GuiPart(master, self.queue)
        self.canv = self.gui.rf.canv  # for easy reference
        
        self.myMCL = MCL(sensorSimulator.Map(self.gui.mapname), NUM_PARTS)
        self.rH = roombaHistory()
        self.mySLAM = SLAM(NUM_PATHS)
        
        # Set up the thread to do asynchronous I/O or other things
        # More can be made if necessary
        self.sim = threading.Thread(target=self.simulator)
        self.sim.setDaemon(1)
        self.sim.start()
        
        try:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.sock.connect((HOST, PORT))
            self.sock.send("kO\n")  # Load color defs
            self.sock.recv(128)
            #self.sock.send("kc\n")  # Begin connected components
            #self.sock.recv(128)
        except:
            print "No vision.\n"
        
        self.keyFeature = 0
        
        # Start the periodic call in the GUI to check if the queue contains
        # anything
        self.gui.processIncoming()
        
        
    def simulator(self):
        """
        This is the non-main thread.
        """
        # the robot and its controlling program
        r       = roombaSim.roombaSim(0)#, self.gui.mapname)
        program = controller.Controller() 
        programRunning = False
        historyRunning = False
        
        # old and current values of robot velocities
        tv_old = 0.0; av_old = 0.0
        tv = 0.0;     av = 0.0
        
        [newpos, collisions], sensors = r.getData()    #getUpdatedPose()
        self.rH.addPose(newpos[0], newpos[1], newpos[2], time.time(), collisions, sensors)
        self.gui.hist.addPoint(self.rH.getNewPoint())
        
        # TEST CODE:
        self.gui.hist.Off()
        
        #f = feature([-5,5],[(10,0),(0,30)])
        #self.gui.mostProbableSLAMPath.addFeatures([f])
        #self.gui.mostProbableSLAMPath.addPoint( (0,0) )
        #self.gui.mostProbableSLAMPath.addPoint( (100, 4) )
        # END TEST CODE
        
        # the "sense - plan - act" loop
        while True:  
            
            if not historyRunning:
                # "sense" - get data from robot and display it
                [newpos, collisions], sensors = r.getData()    #getUpdatedPose()
                self.gui.simrobot.setGlobalPose(*newpos)       #update graphics
            
            # special robot events
            if len(collisions) > 0 and not historyRunning:
                self.updateGUIandMCL(collisions, newpos, sensors)
                tv = 0 ; av = 0

            # ---------------------------------------------------------
                
            # "plan" by running our velocity-choosing program (if kbd-enabled)
            if (programRunning):
                tv, av = program.getVels(newpos, collisions, sensors)
                
            # special user events may override the plan
            c = self.canv.lastKeyDown
            if   c == 'i':  tv =  50; av =   0  # ; self.updateGUIandMCL(collisions, newpos)
            elif c == 'k':  tv = -50; av =   0 
            elif c == 'j':  tv =   0; av =  40
            elif c == 'l':  tv =   0; av = -40
            elif c == ' ':  tv =   0; av =   0 ; self.updateGUIandMCL(collisions, newpos, sensors)
            elif c == 'm':  self.updateGUIandMCL(collisions, newpos, sensors)
            elif c == 'u':  self.updateSLAM(newpos, sensors) ; self.canv.lastKeyDown = '!'
            elif c == 'p':  programRunning = True   # safety tip I learned the hard way:
            elif c == 'P':  programRunning = False  # don't try to toggle with a single key!
            elif c == 'h':  self.gui.hist.On()
            elif c == 'H':  self.gui.hist.Off()
            elif c == 's': 
                self.rH.writeToFile("history.txt") 
                self.canv.lastKeyDown = '!'
            elif c == 'o': 
                self.rH.readFromFile("history.txt")
                self.canv.lastKeyDown = '!'
                self.gui.hist.Off()
                self.gui.hist = RobotCanvas.MovablePath(self.gui.rf.canv, [], colorstr='blue')
                for i in range(len(self.rH.path.x)):
                    self.gui.hist.addPoint(self.rH.getNextTimeStep())
            elif c == 'E': 
                self.rH.readFromFile("history.txt")
                self.canv.lastKeyDown = '!'
                self.gui.hist.Off()
                self.gui.hist = RobotCanvas.MovablePath(self.gui.rf.canv, [], colorstr='blue')
                historyRunning = True
                print "Entering history mode..."
                self.myMCL = MCL(sensorSimulator.Map(self.gui.mapname), NUM_PARTS)
            elif c == 'e':
                if historyRunning:
                    try: 
                        self.gui.hist.addPoint(self.rH.getNextTimeStep())
                        collisions, newpos, sensors = self.rH.getDataNextTimeStep()
                        self.gui.simrobot.setGlobalPose(*newpos)
                        self.myMCL.updateMCL(collisions, newpos)
                        for i in range(NUM_PARTS):
                            self.gui.parts[i].setGlobalPose(*self.myMCL.pose(i))
                    except:
                        print "No more steps."
                self.canv.lastKeyDown = '!'
            elif c == '1':  self.keyFeature = 1; self.updateSLAM(newpos, sensors) ; self.canv.lastKeyDown = '!'
            elif c == '2':  self.keyFeature = 2; self.updateSLAM(newpos, sensors) ; self.canv.lastKeyDown = '!'
            elif c == '3':  self.keyFeature = 3; self.updateSLAM(newpos, sensors) ; self.canv.lastKeyDown = '!'
            elif c == '4':  self.keyFeature = 4; self.updateSLAM(newpos, sensors) ; self.canv.lastKeyDown = '!'
            elif c == '5':  self.keyFeature = 5; self.updateSLAM(newpos, sensors) ; self.canv.lastKeyDown = '!'
            elif c == '6':  self.keyFeature = 6; self.updateSLAM(newpos, sensors) ; self.canv.lastKeyDown = '!'
            elif c == 'q':  break
            
            # ---------------------------------------------------------
            
            # now "act" by setting velocities
            if tv != tv_old or av != av_old:
                r.setVels(tv,av)
                tv_old = tv ; av_old = av
            
            # maybe we should not even have a pause... isn't that what the
            # operating system is for? Or maybe by allowing a pause, things will
            # be smoother... ? Or maybe it's needed for other reasons
            time.sleep(0.025) # 40 hz
               
        
        self.sock.close()
        # if the above while loop ends, we shutdown the gui, too. 
        self.gui.shutdown()
        
        
    def updateGUIandMCL(self, collisions, newpos, sensors):
        #self.myMCL.updateMCL(collisions, newpos)
        #for i in range(NUM_PARTS):
        #    self.gui.parts[i].setGlobalPose(*self.myMCL.pose(i))
        self.rH.addPose(newpos[0], newpos[1], newpos[2], time.time(), collisions, sensors)
        self.gui.hist.addPoint(self.rH.getNewPoint())
 
    def updateSLAM(self, newpos, sensors):
        #j, z = self.getVision()
        try:
            sensorsWithVision = sensors, self.getVision()
            self.mySLAM.updateSLAMkdc(newpos, sensorsWithVision)
        except:
            pass
        try:
            mpp = self.mySLAM.getMostProbablePath()
            self.gui.mostProbableSLAMPath.delete()
            self.gui.mostProbableSLAMPath = RobotCanvas.MovableSLAMPath(self.gui.rf.canv, [], colorstr='orange')
            #for f in mpp.features:
                #if f.seen:
                    #print f.covar
            #print newpos
            for i in range(len(mpp.path.x)):
                self.gui.mostProbableSLAMPath.addPoint( (mpp.path.x[i], mpp.path.y[i]) )
            self.gui.mostProbableSLAMPath.addFeatures(mpp.features)
        except:
            print "Error in paths."
    
    def getVision(self):
        #talk to vision :-D
        #needs to return a RANGE SENSE (r, phi) robot relative
		# and for known correspondence a feature number
        if self.keyFeature == 6:
            r = input('type reading: ')
            self.keyFeature = 0
            return r
        elif self.keyFeature > 0:
            r = self.keyFeature, [25, math.radians(0)], 1
            self.keyFeature = 0
            return r
        try:
            print "talking to vision..."
            self.sock.send("update\n")
            n, x, y = map(int, self.sock.recv(128).split())
            print n, x, y
            y -= 480;
            y = math.fabs(y)
            x -= 320;
            return n, [25, 0]
        except:
            print "No feature."
            return -1
        
root = Tkinter.Tk()
root.withdraw()
toplevel1 = Tkinter.Toplevel(master=root)
#toplevel2 = Tkinter.Toplevel(master=root)
client1 = ThreadedClient(toplevel1)
#client2 = ThreadedClient(toplevel2)

try:
    root.mainloop()
except KeyboardInterrupt:
    print 'Control-c caught from gui...'
    
sys.exit(0)