#
# Ben Tribelhorn
# 7-10-06
#
# High-level driver for the roomba, allows for straight line driving and
# threaded polling of the sensors for bump detection.
#

import roomba
import threading
from time import sleep, time
import sys
from math import fabs, pi
import random

class RoombaPlus:
	def __init__(self, port='/dev/cu.RooTooth-COM0-1'):
		"""
		Initializes a roomba, tests the connection, and instantiates
		 a thread for polling the robot for bumps.
		"""
		print 'Trying connection...'
		self.r = roomba.RoombaRobot(port)
		sleep(1.0)
		self.r.roomba.start()
		self.r.roomba.toSafeMode()
		self.r.roomba.toFullMode()
		self.LINE_DRIVING = False
		self.DRIVING = True
		self.left = self.right = self.bumps = False
		try:
			tmp = (self.r.roomba.readSensors()).distance
			print 'Connected'
			sleep(0.2)
		except:
			print "Not connected."
			return
		self.lock = threading.Lock()
		self.driver = threading.Thread(target=self.fullDataLoop)#self.translate)
		self.driver.setDaemon(1)
		self.driver.start()
		sleep(0.3)

	def toFullMode(self):
		self.r.roomba.toFullMode()
	
	def drive(self, roomba_mm_sec, roomba_radius_mm, turn_dir='CCW'):
		"""
		Sets the roomba's velocity (mm/s) and radius of curvature (mm).
		"""
		if turn_dir != 'LINE':
			self.lock.acquire()
			self.r.roomba.drive(roomba_mm_sec, roomba_radius_mm, turn_dir)
			#self.lock.notify()
			self.lock.release()
			self.LINE_DRIVING = False
		else:
			self.mm_sec = roomba_mm_sec
			self.rad_sec = 2000
			self.LINE_DRIVING = True
	
	def translate(self):
		alpha = 0.4
		cycletime = 0.2
		lefttime = alpha*cycletime
		righttime = cycletime - lefttime
		while(self.DRIVING):
			if self.LINE_DRIVING == True:
				self.lock.acquire()
				if self.LINE_DRIVING == True:
					self.r.roomba.drive(self.mm_sec, self.rad_sec)
				#self.lock.notify()
				#self.lock.release()
				sleep(lefttime)
			#if self.LINE_DRIVING == True:
				#self.lock.acquire()
				if self.LINE_DRIVING == True:
					self.r.roomba.drive(self.mm_sec, -self.rad_sec)
				self.lock.notify()
				self.lock.release()
			sleep(righttime)
			
	def close(self):
		"""
		Closes the roomba.
		"""
		self.DRIVING = False
		sleep(0.2)
		self.r.roomba.start()
		self.r.roomba.close()
		
	def checkBump(self):
		#while self.DRIVING:
			try:
				#print "getting lock"
				self.lock.acquire()
				#print "reading..."
				s = self.r.roomba.readSensors(1)
				left = s.leftBump
				right = s.rightBump
				if left or right:
					print "stopping"
					self.LINE_DRIVING = False
					self.r.roomba.drive(-100,2000)
					sleep(0.6)
					self.r.roomba.drive(0,0)
					self.left = left
					self.right = right
					self.bumps = True
				else:
					self.bumps = False
				#self.lock.notify()
				#print "releasing"
				self.lock.release()
				sleep(0.025)
			except:
				self.lock.release()
				sleep(0.025)
		
	def getData(self):
		try:
			self.lock.acquire()
			x, y, th, left, right = self.r.GetPoseReal()
			#left = bool(left)
			#right = bool(right)
			#if left or right:
			#	sleep(0.02)
			#	self.r.roomba.drive(0,0)
			#	sleep(0.02)
			#self.lock.notify()
			self.lock.release()
			if self.bumps:
				left = self.left
				right = self.right
				self.bumps = False
			th = float(th)
			while th > pi*2:
				th -= 2*pi
			while th < 0:
				th += 2*pi
			newpos = map(float, [x, y, th])
			sensors = [left, left and right, right]
			collisions = []
			FRONT = 0.44
			LEFT = 1.31
			#need to make collisions the theta of the bump
			if left and right:
				collisions = [random.uniform(-FRONT, FRONT)]
				print "FRONT"
			elif left:
				collisions = [random.uniform(FRONT, LEFT)]
				print "LEFT"
			elif right:
				collisions = [random.uniform(-LEFT, -FRONT)]
				print "RIGHT"
			return [newpos, collisions], sensors
		except:
			#print "No robot data--No robot :-("
			#self.close()
			#sys.exit(0)
			sleep(0.01)
			print "failed data grab-trying again..."
			#self.lock.notify() BEN?
			self.lock.release()
			return self.getData()
			
	def fullDataLoop(self):
		alpha = 0.4
		cycletime = 0.2
		lefttime = alpha*cycletime
		righttime = cycletime - lefttime
		while self.DRIVING:
			#self.lock.acquire()
			self.checkBump()
			if self.LINE_DRIVING == True:
				self.lock.acquire()
				#if self.LINE_DRIVING == True:
				self.r.roomba.drive(self.mm_sec, self.rad_sec)
				#self.lock.notify()
				self.lock.release()
			self.sleepCheck(lefttime)
			if self.LINE_DRIVING == True:
				self.lock.acquire()
				#if self.LINE_DRIVING == True:
				self.r.roomba.drive(self.mm_sec, -self.rad_sec)
				#self.lock.notify()
				self.lock.release()
			self.sleepCheck(righttime)
			
			#self.lock.release()
			sleep(0.025)
			
	def sleepCheck(self, tstep):
		newtime = time() + tstep
		while time() < newtime:
			self.checkBump()
		
	def setVels(self, tv, av):
		"""
		Sets the roomba's translational and angular veolocities.
		"""
		if av >= 0: a = 1
		elif av < 0: a = -1;
		if av == 0 and tv == 0: a = av
		if tv == 0 and av != 0:
			self.drive(fabs(av), a)
		else:
			self.drive(tv, a*2000)#, 'LINE')
		