#
# Ben Tribelhorn
# 6-23-06
#
# An implementation of FastSLAM 1.0 with known data correspondence.
#

from path import Path
import math
import random
import copy
from myMath import rotate
from numarray import matrix		## http://sourceforge.net/projects/numpy/ ##

I = matrix.Matrix([[1,0],[0,1]])

class feature:
	def __init__(self, pose, var, s=True):
		"""
		Takes a global position vector (x, y)
		and a covariance (2x2 matrix).
		Optional arguement is if the feature has been seen.
		"""
		self.x = pose[0]
		self.y = pose[1]
		self.covar = var
		self.seen = s
		
	def firstSeenAt(self, pose, var):
		"""
		Sets a new pose, covariance, and seen to True.
		"""
		self.x = pose[0]
		self.y = pose[1]
		self.covar = var
		self.seen = True

class slamPath:
	def __init__(self, weight):
		self.path = Path()
		self.path.addPose(0.0, 0.0, 0.0)
		self.features = []
		self.w = weight
		
	def addFeature(self, f):
		self.features.append(f)
		
	def initNFeatures(self, n):
		for i in range(n):
			self.addFeature(feature([0.0,0.0,0.0], [(0, 0), (0, 0)], False))
		
	def addPose(self, nx, ny, nthr):
		self.path.addPose(nx, ny, nthr)
		
	def getLastPose(self):
		self.path.getLastPose()
        
	def landmarkIsNew(self, i):
		return not self.features[i].seen
		
class SLAM:
	def __init__(self, numPaths=10, numFeatures=5):
		"""
		Creates the SLAM object.  Sets up particles and landmarks.
		"""
		self.NUM_PATHS = numPaths
		self.PL = self.createPaths(numFeatures)
		#self.oldPose = 0.0, 0.0, 0.0
		
	def createPaths(self, n=5):
		PL = []
		weight = 1.0 / self.NUM_PATHS
		
		for i in range(0, self.NUM_PATHS):
			p = slamPath(weight)
			p.initNFeatures(n)
			PL.append(p)
			
		return PL
		
	def updateSLAMkdc(self, newpos, sensors):
		"""
		FastSLAM 1.0 with known data correspondance.
		"""
		bumps, landmark = sensors
		default_weight = 1.0 / self.NUM_PATHS
		
		if len(landmark)>2:
			sm = landmark[2]
		else:
			sm = 1
		if len(landmark)>3:
			a = landmark[3]
		else:
			a = 0.0
		
		print "new pos: ", newpos	
		newpos = input("enter my position:")
		
		for p in self.PL:
			#sample pose
			#newpos = self.addMotionErr(newpos)
			p.addPose(*newpos)
		
			#update feature est.
			#j is index of feature -> for now assume vision tells us what landmark it is
			j = landmark[0]
			dist = landmark[1]
			zt = matrix.Matrix(dist)
			R = self.scalar_mul(sm, self.initUncertianty(a))
			if p.landmarkIsNew(j):
				#init mean and variance

				u = self.ginverse(zt, newpos)		#init mean
				G = self.gprime(newpos, u)			#calculate Jacobian
				try:
					Gi = G**-1						#invert
				except:
					Gi=G	#total hack for when matrix is singular--shouldn't happen with real values
					print "Error: G is singular."
				covar = Gi*R*matrix.transpose(Gi)	#init covariance
				#print "Initial covariance: ", covar
				p.features[j].firstSeenAt(u,covar)	#set landmark seen
				p.w = default_weight				#default importance weight
				print "Placed at: ", u
			else:
				#update the feature
				
				u = p.features[j].x, p.features[j].y
				zhat = self.g(u, newpos)			#measurement prediction
				G = self.gprime(newpos, u)			#calculate Jacobian
				
				#measurement coavariance:
				Q = G * p.features[j].covar * matrix.transpose(G) + R
				
				try:
					Qi = Q**-1
				except:
					print "Error: Q is singular."
					Qi=Q	#another total hack
				
				#calculate Kalman gain
				K = p.features[j].covar * matrix.transpose(G) * Qi
				
				#update mean
				temp = K * matrix.transpose(zt - zhat)
				print "I am moving landmark by: ", temp
				p.features[j].x = p.features[j].x + temp[0]
				p.features[j].y = p.features[j].y + temp[1]
				print "Landmark is at: ", p.features[j].x, p.features[j].y
				#update covariance
				p.features[j].covar = (I - K * G) * p.features[j].covar
				#print "New coavariance: ", p.features[j].covar
				
				#update weight
				#note: I added the absolute value calls here-->may not be correct
				tpq = 1.0/math.sqrt( math.fabs(self.scalar_mul(2.0*math.pi, Q).determinant()))
				epow = math.exp(-0.5 * math.fabs(matrix.transpose(zt-zhat) * Qi * (zt-zhat)))
				if epow == 0:
					#print "Error: weight is zero"
					epow = 0.0001
				p.w = tpq * epow
				#print "updated weight: ", p.w

		#self.oldPose = newpos
		
		#resample
		self.PL = self.resample()
	
	def resample(self):
		"""
		Normalizes particle weights, resamples, and then sets weights to 1.0
		"""
		newPL = []
		sum = 0.0
		#normalize
		for part in self.PL:
			sum += part.w
		for part in self.PL:
			if sum == 0:
				print part.w
			else:
				part.w /= sum
		
		#resample here
		for i in range(len(self.PL)):
			r = random.uniform(0.0, 1.0)
			thresh = 0.0
			for part in self.PL:
				if r <= thresh+part.w:
					newPL.append( copy.deepcopy( part ))
					break
				thresh += part.w
		
		#set resampled (this is probably optional)
		for part in newPL:
			part.w = 1.0
		if len(newPL) == 0:
			print "Error in resample."
			return self.PL
		return newPL
	
	def oldresample(self):
		#return self.PL
		# we want to get l particles out
		newPL = []
		# need the sum of the "probabilities"
		sum = 0.0
		for part in self.PL:
			sum += part.w
		# that's 100%, now start at 0.5
		deltaThresh = sum/len(self.PL) # amount to raise curThresh by each time
		curThresh = deltaThresh/2.0 # seems like a safe starting point
		curPartSum = 0.0
		curPartIndex = 0
		# two nested while loops
		while curPartIndex < len(self.PL):  # particle loop
			curPartSum += self.PL[curPartIndex].w  # add next particle's weight
			while curPartSum > curThresh:  # sampling/probability loop 
				newPL.append( copy.deepcopy(self.PL[curPartIndex]) ) # new copy
				newPL[-1].w =  1.0
				curThresh += deltaThresh  # invariant: append1 for each! curThresh
			curPartIndex += 1  # next particle
		if len(newPL) == 0:
			print "Error in resample."
			return self.PL
		return newPL

	
	def getMostProbablePath(self):
		""" For drawing purposes...
		"""
		w = self.PL[0].w
		mpp = self.PL[0]
		for p in self.PL:
			if p.w > w: mpp = p ; w = p.w
		return mpp

	def addMotionErr(self, pose):
		"""
		This is here for now, but may want to import it from roomba
		Will add random error to motion.
		"""
		#This is totally wrong and useless
		pose[0] += random.uniform(-3, 3)
		pose[1] += random.uniform(-3, 3)
		return pose
	
	def g(self, mu, p):
		"""
		This calculates the distance of the landmark (r) and its angle
		relative to the robot (phi).
		mu is where the landmark *should* be and p is where the robot
		*thinks* it is.
		"""
		r = math.sqrt((mu[0]-p[0])**2 + (mu[1]-p[1])**2)
		phi = math.atan2(mu[1]-p[1], mu[0]-p[0])-p[2]
		return r, phi
	
	def gprime(self, p, mu):
		"""
		Jacobian of g with respect to landmark location.
		"""
		uminusx = mu[0]-p[0]
		vminusy = mu[1]-p[1]
		r = math.sqrt((uminusx)**2 + (vminusy)**2)
		r2 = r**2
		return matrix.Matrix([[(uminusx)/r,(vminusy)/r], [-(vminusy)/(r2),(uminusx)/(r2)]])
		
	def ginverse(self, z, p):
		"""
		Given a reading (z) and estimated robot position (p) this 
		calculates what the absolute position of the landmark should be.
		"""
		r = z[0]
		phi1 = z[1]
		x = p[0]
		y = p[1]
		phi2 = p[2]
		dx = r * math.cos(phi1)
		dy = r * math.sin(phi1)
		if phi2>math.pi:
			phi2 -= 2*math.pi
		rdx, rdy = rotate([dx, dy], phi2)
		u = x + rdx
		v = y + rdy
		return u, v
		
	def initUncertianty(self, a=0.0):
		"""
		Gaussian random vector to model uncertianty.
		Numbers are in r, phi space.
		"""
		#rotation matrix
		A = matrix.Matrix([(math.cos(a), -math.sin(a)), (math.sin(a), math.cos(a))])
		return matrix.transpose(A)*matrix.Matrix([(15,0),(0,math.radians(15))])*A
	
	def scalar_mul(self, s, m):
		"""
		Assumes square matrix.
		"""
		l = len(m[0:])
		nn = copy.deepcopy(m)
		for i in range(l):
			for j in range(l):
				nn[i][j] *= s
		return nn
	
	def reciprocal(self, m):
		"""
		Assumes square matrix.
		"""
		l = len(m[0:])
		nn = copy.deepcopy(m)
		for i in range(l):
			for j in range(l):
				nn[i][j] = 1.0/nn[i][j]
		return nn


